library(ape)


# This provides access to all other file paths
# It can be overidden by a command line argument
SimulationFpathsFpath = "C:/Users/GRJ/AAA/Programming/SARS-Cov2-2/weibull-fsc-sims/simulation6/GRJsimulationFpaths.txt"


#############################################################################################
####################    Reading inputs: file paths and parameters  ##########################


getfpathsFpath <- function() {
	simfpsfpath = SimulationFpathsFpath
	args = commandArgs(trailingOnly=TRUE)
	if (length(args) > 0) {
		simfpsfpath = args[1]
		stopifnot(length(args)==1 )
	}
	simfpsfpath
}


readFpaths <- function(simfpathsFpath) {
	fpathsTable = read.table(simfpathsFpath, header=TRUE)
	stopifnot(fpathsTable[1,"name"] == "filepathSeparator")
	stopifnot(fpathsTable[2,"name"] == "epiDataLeafName")
	stopifnot(fpathsTable[3,"name"] == "sampleDataLeafName")
	stopifnot(fpathsTable[4,"name"] == "fastSimCoalParametersLeafName")
	stopifnot(fpathsTable[5,"name"] == "BeastXMLLeafName")
	stopifnot(fpathsTable[6,"name"] == "paramsFpath")
	stopifnot(fpathsTable[7,"name"] == "replicateDirBase")
	stopifnot(fpathsTable[8,"name"] == "runBatchFPath")
	stopifnot(fpathsTable[9,"name"] == "fsc26FPath")
	fpaths.names = as.vector(fpathsTable[,"name"])
	fpaths.values = as.vector(fpathsTable[,"value"])
	names(fpaths.values) <- fpaths.names
	fpaths.values
}




readParameters <- function(fpaths.values) {
	paramsTable = read.table(fpaths.values["paramsFpath"], header=TRUE)
	stopifnot(paramsTable[1,"name"] == "nReplicates")
	stopifnot(paramsTable[2,"name"] == "nSusceptibles")
	stopifnot(paramsTable[3,"name"] == "minHosts")
	stopifnot(paramsTable[4,"name"] == "minSamples")
	stopifnot(paramsTable[5,"name"] == "PrHostIsSampled")
	stopifnot(paramsTable[6,"name"] == "UNITSPERDAY")
	stopifnot(paramsTable[7,"name"] == "infMaxDays")
	stopifnot(paramsTable[8,"name"] == "startDaysBeforeSeq")
	stopifnot(paramsTable[9,"name"] == "endDaysAfterSeq")
	stopifnot(paramsTable[10,"name"] == "infShape")
	stopifnot(paramsTable[11,"name"] == "infScale")
	stopifnot(paramsTable[12,"name"] == "infIntensity")
	stopifnot(paramsTable[13,"name"] == "population")
	stopifnot(paramsTable[14,"name"] == "bottleneck")
	stopifnot(paramsTable[15,"name"] == "recombination")
	stopifnot(paramsTable[16,"name"] == "mutation")
	stopifnot(paramsTable[17,"name"] == "mutationBias")
	stopifnot(paramsTable[18,"name"] == "genomeSize")
	stopifnot(paramsTable[19,"name"] == "errRate")
	stopifnot(paramsTable[20,"name"] == "sampleSize")
	stopifnot(paramsTable[21,"name"] == "virtualPop")
	stopifnot(paramsTable[22,"name"] == "sampleEvery")
	stopifnot(paramsTable[23,"name"] == "mcmc")
	stopifnot(paramsTable[24,"name"] == "estimateTheta")
	param.names = as.vector(paramsTable[,"name"])
	param.values = as.vector(paramsTable[,"value"])
	names(param.values) <- param.names
	param.values
}


#############################################################################################
##########################            Simulation        #####################################

infectivity <- function(t, infShape, infScale) {
	dweibull(t, shape=infShape, scale=infScale * UNITSPERDAY)
}


get.sample <- function(host, inf.t, startDaysBeforeSeq, actual) {
	smp.t = round(runif(1, min=inf.t, max=inf.t + startDaysBeforeSeq * UNITSPERDAY))
	act = 0
	if (actual) {
		act = 1
	}
	c(host, 1000*host+1, smp.t, act)
}



run.outbreak <- function(attempt, param.values) {
	
	nSusceptibles = param.values["nSusceptibles"]
	minHosts = param.values["minHosts"]
	minSamples = param.values["minSamples"]
	PrHostIsSampled = param.values["PrHostIsSampled"]
	infMaxDays = param.values["infMaxDays"]
	startDaysBeforeSeq = param.values["startDaysBeforeSeq"]
	endDaysAfterSeq = param.values["endDaysAfterSeq"]
	infShape = param.values["infShape"]
	infScale = param.values["infScale"]
	infIntensity = param.values["infIntensity"]
	
	# time.since.infection[j] < 0 means host j is susceptible
	# Start with all susceptible except host 1
	time.since.infection = rep(-1, nSusceptibles)
	time.since.infection[1] = 0
	
	# There are two subsets of hosts.
	# 1. Unobserved, unsampled hosts which can transmit until param.values["infMaxDays"].
	# 2. Observed and Sampled hosts which cannot transmit param.values["endDaysAfterSeq"] after sampling.
	# All hosts need samples for fsc26, but only the second subset go into the XML 
	to.be.sequenced =  as.logical(rbinom(nSusceptibles, 1, PrHostIsSampled))
	
	# maximum time steps in a single infection
	max.inf.steps = infMaxDays * UNITSPERDAY
	# a generous guess at how long the outbreak might last
	n.steps = (1 +  (1 / infIntensity) * log(nSusceptibles) ) * infMaxDays * UNITSPERDAY
	
	infectiousness.table = rep(0, n.steps)
	infectiousness.table[1:max.inf.steps] = infectivity(1:max.inf.steps, infShape, infScale)
	infectiousness.table = infectiousness.table / sum(infectiousness.table)
	infectiousness.table = infectiousness.table * infIntensity  / nSusceptibles  
	cases.so.far = 1
	
	transmissions = matrix(0, ncol=4, nrow=nSusceptibles)
	colnames(transmissions) = c("hostID", "inf.t", "rec.t", "infectorID")
	transmissions[cases.so.far, ] = c(1, 0, max.inf.steps, 0)
	
	# samples.table[i,] stores sample for host i, unsampled hosts are unfilled
	samples.table = matrix(NA, ncol=4, nrow=nSusceptibles) 
	colnames(samples.table) = c("hostID", "genomeID", "smp.t", "actual")
	smp = get.sample(1, 0, startDaysBeforeSeq, to.be.sequenced[1])
	samples.table[1, ] = smp
	
	# main loop
	for (step in 1:n.steps) {
		new.infections = rep(FALSE, nSusceptibles)
		for (i in 1:nSusceptibles) {
			# decide if i can infect
			i.is.infectious = FALSE
			if (time.since.infection[i] >= 0) {
				i.is.infectious = TRUE
				if (to.be.sequenced[i]) {
					endsat = samples.table[i, "smp.t"] + endDaysAfterSeq
					if (step + time.since.infection[i] > endsat) {
						i.is.infectious = FALSE
					}
				}				
			}
			if (i.is.infectious) {
				for (j in 1:nSusceptibles) {
					# decide if i infects j
					if (time.since.infection[j] < 0) {
						if (infectiousness.table[time.since.infection[i] + 1] > runif(1)) {
							new.infections[j] = TRUE
							cases.so.far = cases.so.far + 1
							transmissions[cases.so.far, ] = c(j, step, step+max.inf.steps, i)
							smp = get.sample(j, step, startDaysBeforeSeq, to.be.sequenced[j])
							samples.table[j, ] = smp
						}
					}
				}
			}
		}
		# increment the time that already-infected hosts have been infected
		for (i in 1:nSusceptibles) {
			if (time.since.infection[i] >= 0) {
				time.since.infection[i] = time.since.infection[i] + 1
			}
		}
		# start the new infections for this step
		for (j in 1:nSusceptibles) {
			if (new.infections[j]) {
				time.since.infection[j] = 0
			}
		}
	}
	total.cases = cases.so.far
	# make samples[,] from samples.table[,]
	samples = matrix(NA, ncol=4, nrow=total.cases)
	colnames(samples) = colnames(samples.table)
	for (host in 1:total.cases) {
		hostID = transmissions[host, "hostID"]
		samples[host, "hostID"] = hostID
		samples[host, "genomeID"] = samples.table[hostID, "genomeID"]
		samples[host, "smp.t"] = samples.table[hostID, "smp.t"]
		samples[host, "actual"] = samples.table[hostID, "actual"]
	}
	nActSmps = sum(samples[, "actual"])
	cat("total.cases", total.cases, "actual samples", nActSmps, "\n")
	list(total.cases=total.cases, actual.samples=nActSmps, transmissions=transmissions[1:total.cases,], 
			 samples=samples, to.be.sequenced=to.be.sequenced)
}

#############################################################################################
############################       Drawing the outbreaks     ################################


plot.diagram <- function(w, h)
{
	plot(NULL, xlim=c(0,w), ylim=c(h,0), axes=FALSE, ylab="", xlab="")
}


plot.parallelogram <- function(v1,v2,v3,...)
{
	polygon(c(v1[1],v2[1],v2[1]+v3[1]-v1[1],v3[1]), c(v1[2],v2[2],v2[2]+v3[2]-v1[2],v3[2]), ...)
}



plot.transmission.tree <- function(transms, param.values) {
	nSusceptibles = param.values["nSusceptibles"]
	max.units = max(transms$samples[, "smp.t"]) + 10 * UNITSPERDAY
	plot.diagram(transms$total.cases+2, max.units+2)
	for (d in 1:max.units) {
		if (d %% (UNITSPERDAY/2) == 0) {
			abline(h=d, col=grey(0.85))
		}
		if (d %% UNITSPERDAY == 0) {
			abline(h=d, col=grey(0.6))
		}
	}
	for (d in 1:max.units) {
		if (d %% UNITSPERDAY == 0) {
			text(x=0.25, y=d, d/UNITSPERDAY, adj=c(1,0.5), cex=0.7)
		}
	}
	#plot.parallelogram(c(1,1), c(nSusceptibles,1), c(1, max.units))
	for (i in 1:nrow(transms$transmissions)) {
		hostID = transms$transmissions[i, "hostID"]
		endsat = transms$transmissions[i, "rec.t"]
		sampled = FALSE
		if (transms$to.be.sequenced[hostID]) {
			endsat = transms$samples[i, "smp.t"] + param.values["endDaysAfterSeq"] * UNITSPERDAY
			sampled = TRUE
		}

		inft = transms$transmissions[i, "inf.t"]
		lines(x=c(i,i), y=c(inft, endsat), lwd=7)
		text(x=i, y=inft, labels=paste0("H", transms$transmissions[i, "hostID"]), adj=c(0.5,-.5), cex=0.8)
		if (sampled) {
			smpt = transms$samples[i, "smp.t"]
			lines(x=c(i-0.3,i+0.3), y=c(smpt,smpt), col="blue", lwd=4)
		}
	}
	for (i in 2:nrow(transms$transmissions)) {
		w = which(transms$transmissions[, "hostID"] == transms$transmissions[i, "infectorID"])
		stopifnot(length(w) == 1)
		inft = transms$transmissions[i, "inf.t"]
		lines(x=c(i,w), y=c(inft, inft), col="red", lwd=2)
	}
}


#############################################################################################
########################       Conversion of outbreak to tree     ###########################
# 29 Apr 2020 no longer used
write.phylo.pdf <- function(transms, treefpath) {
	# This translates the transmissions matrix, rows of (hostID, infection time = inf.t, infectorID)
	# Although both formats are a list of edges, they are quite different
	# The first row in transms[,] is the root, which is not wanted as an edge.
	
	# First, find which edges are external, the tip IDs, and the maximum node ID in transms[,]
	nEdges = nrow(transms) - 1
	isExtEdge = rep(FALSE, nEdges)
	maxID = 0
	tipsIDs = integer(0)
	tip = 1
	for (edg in 1:nEdges) {
		chID = transms[1+edg, "hostID"]
		isExtEdge[edg] = !any(transms[,"infectorID"] == chID)
		if (isExtEdge[edg]) {
			tipsIDs[tip] = chID
			tip = tip + 1
		}
		maxID = max(maxID, chID)
		maxID = max(maxID, transms[1+edg, "infectorID"])
	}
	nTips = sum(isExtEdge)
	# nodeIdx.map[] is the key to translating to phyo format.
	# nodeIdx.map[ID] is an index for a phylo object, where ID is a node ID in  transms[,]
	nodeIdx.map = rep(0, maxID)
	tipIdx = 1
	intIdx = 1
	nodeIDs = integer(0)
	# do root
	nodeIdx.map[transms[1, "hostID"]] = nTips + intIdx
	nodeIDs[intIdx] = transms[1, "hostID"]
	intIdx = intIdx + 1
	# do child of every edge
	for (edg in 1:nEdges) {
		chID = transms[1+edg, "hostID"]
		if (isExtEdge[edg]) {
			nodeIdx.map[chID] = tipIdx
			tipIdx = tipIdx + 1
		} else {
			nodeIdx.map[chID] = nTips + intIdx
			nodeIDs[intIdx] = chID
			intIdx = intIdx + 1 
		}
	}
	# edge is the array of edges for phylo. brlens is the edge lengths.
	# For the latter, need to look up where the ancestor of each edge is, to find its time.
	edge = matrix(0, nrow=nEdges, ncol=2)
	brlens = rep(0, nEdges)
	for (edg in 1:nEdges) {
		edge[edg,1] = nodeIdx.map[transms[1+edg,"infectorID"]]
		edge[edg,2] = nodeIdx.map[transms[1+edg,"hostID"]]
		ancID = transms[1+edg, "infectorID"]
		w = which(transms[, "hostID"] == ancID)
		stopifnot(length(w)==1)
		brlens[edg] = transms[1+edg,"inf.t"] - transms[w,"inf.t"]
	}
	# make the phyo object
	myph <- list(edge = edge, tip.label = paste(tipsIDs), node.label = paste(nodeIDs), edge.length = brlens, Nnode=nEdges+1-nTips)
	class(myph) <- "phylo"
	# write PDF
	pdf(file=treefpath, width=5, height=5)
	plot(myph, show.node.label=TRUE)
	dev.off()
}


#############################################################################################
##############  Outbreak to files for epiSmpsToPar.py and drawings  #########################

output.epi.smps <- function(outbreakinfo, fpaths.values, rep) {
	rep.dpath = paste0(fpaths.values["replicateDirBase"], rep - 1)
	if (!dir.exists(rep.dpath)[1]) {
		dir.create(rep.dpath)
	}
	trfpath = paste0(rep.dpath, fpaths.values["filepathSeparator"], fpaths.values["epiDataLeafName"])
	write.table(outbreakinfo$transmissions, file=trfpath, row.names=FALSE, quote=FALSE)
	smpfpath = paste0(rep.dpath, fpaths.values["filepathSeparator"], fpaths.values["sampleDataLeafName"])
	write.table(outbreakinfo$samples, file=smpfpath, row.names=FALSE, quote=FALSE)
	plot.transmission.tree(outbreakinfo, param.values)
}



#############################################################################################
###################################      Top level      #####################################
#############################################################################################

#############################################################################################
#####################         Run simulations, output results           #####################

fpaths.values = readFpaths(getfpathsFpath())
param.values = readParameters(fpaths.values)

UNITSPERDAY = param.values["UNITSPERDAY"]

all.outbreaks = NULL
for (rep in 1:param.values["nReplicates"]) {
	minHosts = param.values["minHosts"]
	minSamples = param.values["minSamples"]
	outbreakinfo = NULL
	for (i in 1:1000) {
		set.seed(1000*rep + i)
		print(paste("Replication", rep, ", attempt", i))
		outbreakinfo = run.outbreak(i, param.values)
		if (outbreakinfo$total.cases >= minHosts  &&  outbreakinfo$actual.samples >= minSamples) {
			break
		} else {
			outbreakinfo = NULL
		}
	}

	all.outbreaks = c(all.outbreaks, list(outbreakinfo))
}

diagfpath = paste0(fpaths.values["replicateDirBase"], "s_diags.pdf")
pdf(file=diagfpath)
for (b in 1:length(all.outbreaks)) {
	ob = all.outbreaks[[b]]
	if (!is.null(ob)) {
		output.epi.smps(ob, fpaths.values, b)
	} else {
		stop(paste("Replication", rep, "failed to reach", minHosts, "hosts and ", minSamples, "samples"))
	}	
}
dev.off()


#############################################################################################
##############  Make batch file or shell script for running fsc26 later #####################

	
if (.Platform$OS.type == "windows") {
	allrepsDir = paste0(fpaths.values["replicateDirBase"], "s-src-cfg")
	if (!dir.exists(allrepsDir)[1]) {
		dir.create(allrepsDir)
	}
	thisFile = fpaths.values["runBatchFPath"]
	parts = strsplit(thisFile, "\\", fixed=TRUE)[[1]]
	containingDir = parts[length(parts) -1]
	batchLines = c(
		paste0("SET SIMFPATHSFPATH=", getfpathsFpath()),
		"SET BATCHFILEDIR=%cd% ",
		"REM copy folder containing this batch file to output (R and Python source, input params)",
		paste0("XCOPY ..\\", containingDir, " ", allrepsDir),
		"",
		paste0("SET FSC26PATH=", fpaths.values["fsc26FPath"]),
		paste0("SET PYTHON=C:\\Users\\GRJ\\AAA\\software\\AnaConda\\python"),
		"",
		"%PYTHON% epiSmpsToPar.py %SIMFPATHSFPATH%",
		"IF ERRORLEVEL 1 PAUSE",
		""
	)
	
	for (rep in 1:param.values["nReplicates"]) {
		repDir = paste0(fpaths.values["replicateDirBase"], rep-1)
		batchLines = c(batchLines,
									 paste0("CD ", repDir),
									 paste0("%FSC26PATH% -i ", fpaths.values["fastSimCoalParametersLeafName"], " -n 1 --seed ", rep-1, " > fsc26_stdout.txt"),
									 "IF ERRORLEVEL 1 PAUSE",
									 ""
		)
	}
	batchLines = c(batchLines, 
								 "",
								 "CD %BATCHFILEDIR%",
								 "%PYTHON% make-BADTRIP-XML.py %SIMFPATHSFPATH%",
								 "IF ERRORLEVEL 1 PAUSE",
								 ""
	)
	batchLines = c(batchLines,"PAUSE", "")
}
if (.Platform$OS.type == "unix") {
	
	
	# TODO
	
	
}
fpath = fpaths.values["runBatchFPath"]
fileConn<-file(fpath)
open(fileConn, "w")
writeLines(batchLines, fileConn)
close(fileConn)






